# Copyright (c) 2013-2014  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import scum.scanner
import collections

from scum.scanner import eoi as eoi_, start as start_

class Token(object):
    r""" This class represents a token from an input stream. A token
    has a value, type identifier and the cursor position it was found
    at. When the end of an input stream was reached, a Token with a
    None value is created by the :class:`Lexer`. Checking for the
    boolean value of the Token yields False in that special case.

    If the :class:`Lexer` is configured to also yield at characters
    that could not be gathered to a Token, the Token may contain a
    single-character string as its value only and the identifier is
    None.

    Additional attributes of a Token can be defined by passing keyword
    arguments to the constructor.

    .. attribute:: id

        The identifier of the Token which can be any object that was
        defined to represent the type of the Token. Numbers or strings
        are very common.

        If this attribute is None, the Token is an un-identified Token
        which was not declared in any of the token type declarations or
        if the Token represents the end of the input stream in which
        case the :attr:`value` is None, too.

    .. attribute:: value

        The value of the Token, a string unless the Token marks the end
        of the input stream.

    .. attribute:: cursor

        The cursor position the Token was found at in the input stream.
        A :class:`scum.scanner.Cursor` instance.
    """

    def __init__(self, token_id, value, cursor, attrs={}):
        assert isinstance(cursor, scum.scanner.Cursor)

        super(Token, self).__init__()
        self.id = token_id
        self.value = value
        self.cursor = cursor.copy()
        self.attrs = dict(attrs)

    def __nonzero__(self):
        return bool(self.value)

    def __repr__(self):
        name = self.__class__.__name__
        if not self.value:
            return '<%s at eoi>' % name

        value = str(self.value)
        if len(value) > 20:
            value = value[:9] + '...' + value[9:]

        result = '<%s(%s) %r at %s' % (name, self.id, value, self.cursor)
        if self.attrs:
            result += ' %s' % self.attrs
        return result + '>'

    def __getattr__(self, name):
        try:
            return self.attrs[name]
        except KeyError:
            raise AttributeError(name)

class TokenExtractor(object):
    r""" The base class for objects that are capable of extracting
    a :class:`Token` from a :class:`scum.scanner.Scanner`. """

    def extract_token(self, scanner, token_id, cursor, token_attrs):
        r""" Called to extract and return a :class:`Token` from
        the *scanner*. Return None to tell that no Token could be
        extracted. """

        raise NotImplementedError

class TokenExtractorWrapper(TokenExtractor):
    r""" A wrappr for callable objects that accept only a Scanner
    as their first and only argument. """

    def __init__(self, func):
        super(TokenExtractorWrapper, self).__init__()
        self.func = func

    def extract_token(self, scanner, token_id, cursor, token_attrs):
        result = self.func(scanner)
        if result:
            return Token(token_id, result, cursor, token_attrs)
        else:
            return None

class Lexer(object):
    r""" The Lexer uses the :class:`scum.scanner.Scanner` class to
    analyze an input character stream and split it into Tokens which
    are declared via :meth:`declare_token`.

    A Lexer is re-usable as the Scanner can be plugged in and out. """

    def __init__(self):
        super(Lexer, self).__init__()
        self.__tokendefs = []
        self.__token = None
        self.__scanner = None

    def __iter__(self):
        if not self.__scanner:
            raise RuntimeError('Lexer has not source assigned')

        while self.next():
            yield self.__token

    @property
    def token(self):
        r""" Returns the last Token that was extracted from the input
        stream. The ID of the Token is None if no Token declaration
        was able to process the value of the Token (which is a one-
        character string). If the value of the Token is None, the end
        of the input stream was reached. """

        return self.__token

    @property
    def is_initialized(self):
        r""" Returns True if the Lexer is initialized and lexing is
        in progress (that is, when a Scanner was set as the source of
        the Lexer via :meth:`set_source`). """

        return bool(self.__scanner)

    def declare_token(self, token_id, extractor, **token_attrs):
        r""" Declares a Token to the Lexer. The order of declaration
        defines the order they are processed, the first declared Token
        is applied first. The first matching declaration is used as the
        actual Token. If *skip* is True, the Token will be skipped by
        the Lexer and not even yielded but instead the next Token is
        read.

        Two declarations with the same ID are possible in case that
        is of any use for someone.

        Passing None for the *token_id* will cause the Lexer to
        skip the Token completely when it was extracted and instead
        extract the next Token.

        The *extractor* argument must be either a callable object or an
        instance of the :class:`TokenExtractor` class. If it is a
        callable object, it must accept a :class:`scum.scanner.Scanner`
        as its only argument and return a string of the extracted Token
        value.
        Both, a callable object or a TokenExtractor, do not need to
        take care of re-setting the Scanners cursor to the original
        position.

        For extended usage, Token attributes can be specified by
        passing additional keyword arguments which will be set as
        attributes for the Token generated by this extractor.

        :raise RuntimeError: If the Lexer has a source already
            assigned (see :attr:`is_initialized`). """

        if self.__scanner:
            raise RuntimeError('can not declare new tokens while lexing is '
                               'in progress')

        if not isinstance(extractor, TokenExtractor):
            if isinstance(extractor, collections.Callable):
                extractor = TokenExtractorWrapper(extractor)
            else:
                raise TypeError('extractor is not callable nor TokenExtractor')

        self.__tokendefs.append((token_id, extractor, token_attrs))

    def set_source(self, scanner):
        r""" Set the source of the input stream to a
        :class:`scum.scanner.Scanner` instance. Raises TypeError if
        *scanner* is not such an instance.

        After this method was successfully called,
        :meth:`declare_token` can not be called until the Lexer
        was de-initialized with :meth:`unplug_source`.

        A RuntimeError is raised when this method is called multiple
        times without de-initializing the Lexer beforehand. """

        if self.__scanner:
            raise RuntimeError('can not overwrite existing source of lexer')
        if not isinstance(scanner, scum.scanner.Scanner):
            raise TypeError('scanner must be scum.scanner.Scanner instance')

        self.__scanner = scanner
        if scanner.char == start_:
            scanner.next()

    def unplug_source(self):
        r""" Unplugs the source of the Lexer and de-initializes it.
        Raises a RuntimeError when called in an un-initialized state. """

        if not self.__scanner:
            raise RuntimeError('Lexer is already un-initialized')

        self.__scanner = None
        self.__token = None

    def next(self):
        r""" Reads the next Token from the input stream. The last
        extracted Token can be accessed via the :attr:`token`
        attribute.

        Returns a :class:`Token` object. If the ID of the returned
        Token is None, no Token could be extracted and it contains
        a one-character string value.
        If the value of the returned Token is None, the end of the
        input stream was reached.

        Raises a RuntimeError if the Lexer is not initialized. """

        scanner = self.__scanner
        if not scanner:
            raise RuntimeError('Lexer has no source assigned')

        # Get the current position of the Scanner as we need
        # to assign it to the extracted Token.
        cursor = scanner.cursor

        # Yield the Token that represents the end of the input
        # stream if we actually are at the end of the stream.
        if scanner.char == eoi_:
            self.__token = Token(None, None, cursor)
            return self.__token

        # This will hold the extracted Token, if any could be
        # extracted.
        token = None

        # Try all declared Token matchers.
        for token_id, extractor, token_attrs in self.__tokendefs:
            result = extractor.extract_token(
                    scanner, token_id, cursor, token_attrs)

            # If an object was returned, it must be a Token instance.
            if result and not isinstance(result, Token):
                message = 'token declaration %r returned non-Token instance %s'
                raise TypeError(message % (token_id, result.__class__.__name__))
            elif result:
                token = result
                break

            # Re-set the Scanners cursor to its original position.
            scanner.cursor = cursor

        # If no Token matched, we will return an invalid
        # un-identified Token.
        if token is None:
            char = scanner.char
            scanner.next()
            token = Token(None, char, cursor)

        # Or should we actually skip the current Token? This is
        # the case when the token_id is None.
        else:
            if token_id is None:
                return self.next()

        self.__token = token
        return token

    def get_and_next(self):
        r""" Returns the current Token and extracts the next. """

        try:
            return self.__token
        finally:
            self.next()

