# Copyright (c) 2013-2014  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
r"""
    scum.matcher
    ~~~~~~~~~~~~

    This module defines many useful matchers which can be used in
    combination with the :class:`scum.scanner.Scanner` class. They
    generally however do not re-set the Scanner to its original
    cursor position if they did not match but just return True as it
    is the definition of matchers by the :class:`scum.lexer.Lexer`
    class.
"""

import string

id_start = string.letters + '_'
id_continue = id_start + string.digits
whitespace = string.whitespace

class Keyword(object):
    r""" This matcher for a :class:`Scanner` matches a specific keyword
    as it is defined with the objects construction. It supporst to
    ignore the case of the letters and also to only match when certain
    characters do not follow the matched keyword. """

    def __init__(self, word, ignore_case=False, verify_word=False,
                 no_follow_set=None):
        r""" Initializes the Keyword matcher.

        :param word: The word to match.
        :param ignore_case: True if the case of the letters should
            be ignored, False if not.
        :param verify_word: If this is True, the *no_follow_set*
            will get a default value of all letters including the
            underscore and digits (see :data:`id_continue`).
        :param no_follow_set: If specified, it must be a set of
            characters that must not appear after the keyword matched. """

        if no_follow_set is None and verify_word:
            no_follow_set = id_continue

        super(Keyword, self).__init__()
        self.word = word
        self.ignore_case = ignore_case
        self.no_follow_set = no_follow_set

    def __call__(self, scanner):
        word = self.word
        if self.ignore_case:
            word = word.lower()

        result = scanner.match(self.word, self.ignore_case, fallback=False)

        # Let's check if the current character of the Scanner is
        # not in the *no_follow_set*. If it is, the Keyword does
        # not match.
        if self.no_follow_set is not None:
            if scanner.char in self.no_follow_set:
                return None

        return result

class Gather(object):
    r""" This matcher gathers all characters from a specified character
    set. A start set can be defined to define the characters the Token
    has to begin with. If the continue set is defined, all other
    characters may be of that other set. """

    def __init__(self, start_set, continue_set=None):
        if continue_set is None:
            continue_set = start_set

        super(Gather, self).__init__()
        self.start_set = start_set
        self.continue_set = continue_set

    def __call__(self, scanner):
        char = scanner.char
        if char not in self.start_set:
            return None

        scanner.next()
        result = char + scanner.read_set(self.continue_set)
        return result

class SkipFrom(object):
    r""" This matcher skips all characters when a certain character
    sequence was found until a certain other character sequence was
    found. """

    def __init__(self, from_, to='\n', ignore_case=False, include_end=True):
        super(SkipFrom, self).__init__()
        self.from_ = from_
        self.to = to
        self.ignore_case = ignore_case
        self.include_end = include_end

    def __call__(self, scanner):
        result = scanner.match(self.from_, self.ignore_case)
        if not result:
            return None
        while True:
            cursor = scanner.cursor
            end = scanner.match(self.to, self.ignore_case, fallback=False)
            if end:
                if self.include_end:
                    result += end
                else:
                    scanner.cursor = cursor
                break

            scanner.cursor = cursor
            result += scanner.char
            scanner.next()

        return result


