# Copyright (c) 2013-2014  Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import collections
from cStringIO import StringIO

# This singleton represents the start of an input stream.
start = type('start_type', (), {
    '__nonzero__': lambda s: False,
    '__repr__': lambda s: '<start>',
})()

# This singleton represents the end of an input stream.
eoi = type('eoi_type', (), {
    '__nonzero__': lambda s: False,
    '__repr__': lambda s: '<eoi>',
})()

class Cursor(object):
    r""" This class is a set of the position, line and column
    number information of the cursor in a :class:`Scanner`.

    .. attribute:: position

        The index in the input stream. This value is decisive when
        seeking in the Scanner.

    .. attribute:: line

        The line-number respective to the position of the cursor.

    .. attribute:: column

        The column-number respective to the position of the cursor.
    """

    def __init__(self, position=0, line=1, column=0):
        super(Cursor, self).__init__()
        self.position = int(position)
        self.line = int(line)
        self.column = int(column)

    def __repr__(self):
        return '<Cursor at %d (line: %d, column: %d)>' % (
                self.position, self.line, self.column)

    def __eq__(self, other):
        if not isinstance(other, Cursor):
            return False
        return (self.position == other.position and self.line == other.line
                and self.column == self.column)

    def __ne__(self, other):
        return not (self == other)

    def copy(self):
        r""" Returns a copy of the Cursor. """

        return Cursor(self.position, self.line, self.column)

class Scanner(object):
    r""" This class is a tool for analyzing input streams of
    characters. Features include

    * Lookback and lookforward in the input stream
    * Line and column counting
    * Convenient data-gathering

    About lines and columns
    ^^^^^^^^^^^^^^^^^^^^^^^

    The line number starts at one while the column number starts at
    zero. When the very first character is read from the input stream,
    assuming it is not a newline character, the column number will be
    one.

    You can imagine the cursor of the Scanner like in a text editor.
    The current character of the Scanner (see :attr:`char`) is always
    the character to the right of the cursor bar.

    The character(-sequence) that is regarded as the newline can
    be modified, but it defaults to the line-feed ``'\n'``. Python
    can open files and automatically universally convert all line
    endings corresponding to the current platform to line-feed
    characters.

    If a newline character(-sequence) is detected, it is yielded
    as the object defined via :attr:`newline_rep`. """

    @classmethod
    def from_string(cls, string, *args, **kwargs):
        r""" Construct a Scanner from a string. """

        fp = StringIO(string)
        return cls(fp, *args, **kwargs)

    def __init__(self, fp, buffersize=512, newline='\n', newline_rep='\n'):
        r""" Construct a new Scanner object. The values of the
        specified parameters can not be changed after construction.

        :param fp: A file-like object that serves as the input stream.
        :param buffersize: The maxmimum number of cached characters.
        :param newline: The newline character(-sequence).
        :param newline_rep: The character that represents a newline
            during the scanning process. If a sequence as defined with
            the :attr:`newline` is found in the input stream, it will
            be represented by this character during the scanning
            process.

            Keep in mind that, for instance, the set provided to
            :meth:`skip_set` must support this object with its
            ``__contains__()`` method.
        :raise TypeError: If *fp* is not a file-like object,
            *newline* is not a :class:`basestring` instance or
            *buffersize* is smaller than the newline sequences length.
        """

        if not hasattr(fp, 'read'):
            raise TypeError('expected file-like object for fp parameter')
        if not isinstance(newline, basestring):
            raise TypeError('expected basestring for newline parameter')
        buffersize = int(buffersize)
        if buffersize <= 0:
            raise ValueError('buffersize must be a positive non-zero value')
        if buffersize < len(newline):
            raise ValueError('buffersize must be at least as large as the '
                    'length of the newline character-sequence')

        # Add one to the buffer size as :data:`start` is included
        # as the first element in the buffer (representing the
        # start of the input stream.
        buffersize += 1

        super(Scanner, self).__init__()
        self.__fp = fp
        self.__buffersize = buffersize
        self.__buffer = collections.deque((start, ))
        self.__char = start
        self.__index = -1
        self.__cursor = Cursor()
        self.__newline = newline
        self.__newline_rep = newline_rep

    def __iter__(self):
        while self.next():
            yield self.__char

    @property
    def fp(self):
        r""" The internal file-object that is used as an input stream
        for the data that is yielded by the Scanner. """

        return self.__fp

    @property
    def buffersize(self):
        r""" This value represents the maximum number of cached
        characters which are required for the lookback and
        lookforward. """

        return self.__buffersize

    @property
    def cursor(self):
        r""" This property returns a :class:`Cursor` object that
        contains the information about line and column number as well
        as the current position of the Scanner in the file. """

        return self.__cursor.copy()

    @cursor.setter
    def cursor(self, cursor):
        r""" Move the cursor of the Scanner to the position designated
        in the specified :class:`Cursor` object. Only cursor objects
        that were retreived by the :attr:`cursor` attribute should be
        used for this method as it is otherwise possible that line
        and column number will not match the actual input stream.

        :raise ValueError: If seeking to the designated location is not
            possible (eg. when the buffersize was not chosen large
            enough). """

        if not isinstance(cursor, Cursor):
            raise TypeError('expected Cursor instance')

        delta = self.__cursor.position - cursor.position
        new_index = self.__index - delta

        # Is the new index located in the buffer? If it is not,
        # we can not go back to this location.
        if new_index < 0 or new_index >= len(self.__buffer):
            raise ValueError('can not seek to location %d' % new_index)

        self.__index = new_index
        self.__cursor = cursor.copy()

        # Read the current character from the buffer.
        self.__char = self.__buffer[new_index]

    @property
    def char(self):
        r""" The current character from the input stream. This is the
        character to the right of the cursor bar (if you would imagine
        the Scanner to be a cursor in a text editor).

        If not a single character has yet been read from the input
        stream, :attr:`char` is :data:`start`. In contrast, if the end
        of the input stream was hit, :attr:`char` is :data:`eoi`. """

        return self.__char

    @property
    def newline(self):
        r""" The newline character(-sequence) of the input source. It
        is specified during construction of the Scanner and can not be
        changed afterwards. """

        return self.__newline

    @property
    def newline_rep(self):
        r""" The character that represents the newline in the input
        stream as a newline while scanning. The default value is the
        line-feed charater. """

        return self.__newline_rep

    def __nextchar(self):
        r""" Reads the next character from the cache or the input
        stream. This is a private method used by :meth:`next`. It
        also increases the cursor position. """

        imax = len(self.__buffer) - 1

        # If the index is pointing right to the last element
        # of the buffer or no character has yet been read, the next
        # character should be read from the input stream and put into
        # the buffer. The size of the buffer must not exceed the limit.
        if self.__index < 0 or self.__index == imax:

            # The buffer must be empty when the index is a negative
            # number.
            if self.__index < 0:
                assert len(self.__buffer) == 1
                assert self.__buffer[0] == start

            # Do not read from the input stream if we have hit
            # the EOF already.
            if self.__buffer[-1] == eoi:
                return eoi

            # Read the character from the input stream and check
            # if it should be treated as a newline.
            char = self.__fp.read(1) or eoi
            if char == self.__newline[0]:
                if self.__match_newline():
                    char = self.__newline_rep

            # Add it to the buffer.
            self.__buffer.append(char)

            # Shrink the buffer from the left until it fits
            # with the defined maximum size.
            while len(self.__buffer) > self.__buffersize:
                self.__buffer.popleft()

            self.__index = len(self.__buffer) - 1

        # If the index is not at the end of the buffer, the
        # next character should be retreived from the buffer
        # and not from the input stream. The index will increase
        # by one.
        elif self.__index < imax:
            self.__index += 1
            char = self.__buffer[self.__index]

        # The index must not exceed the real size of the buffer
        # which is why we will raise an error in this case.
        else:
            raise RuntimeError('something went wrong, the buffer cursor '
                    'index exceeds the size of the buffer')

        # Increase the cursor position if we have a valid character.
        if char:
            self.__cursor.position += 1

        return char

    def __match_newline(self):
        r""" Private method. Called from :meth:`__nextchar` in case
        the first character of the newline sequence was read. Returns
        True if it really matches with the newline character sequence,
        False if it does not. The characters that are read from the
        input file are put on the buffer if they did not match with the
        newline character. """

        matches = True

        # Collects all characters that are being read from the
        # input stream in case it does not match the newline
        # character sequence and they need to be pushed to the
        # buffer.
        chars = collections.deque()

        # This method is only called when the first character
        # did match already so we just check the remaining ones.
        for c in self.__newline[1:]:
            # Get the next character from the input stream and
            # put it into the chars list.
            current = self.__fp.read(1)
            if current:
                chars.append(current)

            # Stop if the current character does not match the
            # required character from the newline sequence.
            if current != char:
                matches = False
                break

        # Put all the characters we've read onto the buffer if
        # it did not match. It's not too bad if the buffer now
        # gets bigger than the buffer size, it'll shrink in the
        # next call to :meth:`__nextchar`.
        if not matches:
            self.__buffer.extend(chars)

        return matches

    def next(self):
        r""" Reads the next character from the input stream and
        returns it. If the buffer is already full, the "oldest"
        character is popped from the front and the new one is
        inserted at the front.

        :return: The character or an empty string if the end of
            the input stream was reached.
        """

        char = self.__nextchar()
        if char == self.__newline_rep:
            self.__cursor.line += 1
            self.__cursor.column = 0
        else:
            self.__cursor.column += 1

        self.__char = char
        return char

    def skip(self, n):
        r""" Skips the next *n* characters and returns the number of
        skipped characters. The returned value can be lower than *n*
        if the end of the input stream is reached in the meantime. """

        if self.__char == start:
            raise RuntimeError('skip() includes the current character, '
                    'Scanner must not be located at <start>')

        count = 0
        for i in xrange(n):
            if self.__char:
                count += 1
            self.next()

        return count

    def read(self, n):
        r""" Reads *n* characters from the input stream including
        the current character and returns a string. The Scanners cursor
        is moved by the specified amount.

        The returned string can be shorter than *n* if the end of
        the input stream was reached. """

        if self.__char == start:
            raise RuntimeError('read() includes the current character, '
                    'Scanner must not be located at <start>')

        result = ''
        for i in xrange(n):
            if not self.__char:
                break
            result += self.__char
            self.next()

        return result

    def skip_set(self, charset, invert=False, max_=-1):
        r""" Skips all the characters from the current position of
        the Scanner that are contained in the *charset*.

        .. note:: This method should only be called when at least
            a single character was read from the Scanner.

        :param charset: A set of characters that provides the
            ``__contains__()`` interface.
        :param invert: Instead of skipping all characters that are
            contained in the *charset*, skip all characters that are
            not in the *charset*.
        :param max_: If this is not a negative number, limits the
            number of characters that are skipped.
        :return: The number of skipped characters. """

        count = 0
        while self.__char:
            if max_ >= 0 and count >= max_:
                break

            contained = self.__char in charset
            if invert: contained = not contained
            if not contained:
                break

            count += 1
            self.next()

        return count

    def read_set(self, charset, invert=False, max_=-1):
        r""" Reads all the characters from the current position of
        the Scanner that are contained in the *charset* and puts
        them into a string.

        .. note:: This method should only be called when at least
            a single character was read from the Scanner.

        :param charset: A set of characters that provides the
            ``__contains__()`` interface.
        :param invert: Instead of reading all characters that are
            contained in the *charset*, read all characters that are
            not in the *charset*.
        :param max_: If this is not a negative number, limits the
            number of characters that are read.
        :return: The string of the read characters. """

        result = ''
        while self.__char:
            if max_ >= 0 and len(result) >= max_:
                break

            contained = self.__char in charset
            if invert: contained = not contained
            if not contained:
                break

            result += self.__char
            self.next()

        return result

    def skip_until(self, charset, max_=-1):
        r""" Skips all characters until either *max_* characters have
        been skipped or a character from *charset* was found. Returns
        the number of skipped characters. """

        return self.skip_set(charset, True, max_)

    def read_until(self, charset, max_=-1):
        r""" Reads all characters until either *max_* characters have
        been read or a character from *charset* was found. Returns the
        string of the read characters. """

        return self.read_set(charset, True, max_)

    def look_back(self, n):
        r""" Returns a string of the *n* last characters. The returned
        string can be empty if there are no characters buffered or the
        start of the input stream was reached. The cursor of the Scanner
        is not moved. """

        result = ''
        for i in xrange(n):
            i = self.__index - 1
            if i < 0: break
            char = self.__buffer[i]
            if not char: break
            result = char + result

        return result

    def look_forward(self, n):
        r""" Returns a string of the *n* following characters. The
        returned string can be empty if the end of the input stream
        was already reached. *n* must not be larger than the specified
        :attr:`buffersize` as the Scanners cursor is not moved with
        this method. """

        if n >= self.__buffersize:
            raise ValueError('n must be smaller than the buffersize')

        cursor = self.cursor
        result = ''
        for i in xrange(n):
            char = self.next()
            if not char: break
            result += char

        self.cursor = cursor
        return result

    def match(self, iterable, ignore_case=False, fallback=True):
        r""" Matches the character sequence specified in *iterable*
        starting with the current character of the Scanner. The
        case of the letters can be ignored by passing True for the
        *ignore_case* parameter. The cursor of the Scanner will be
        moved to its original position when it didn't match unless
        *fallback* is False.

        Returns the string that matched or None if it didn't match. """

        if fallback:
            cursor = self.cursor

        result = ''
        for c in iterable:

            # Read the next character from the Scanner and add it
            # to the resulting string if it is a valid character (and
            # not the eoi).
            o = self.__char
            if o:
                result += o

                # Lower it if this is desired.
                if ignore_case:
                    o = o.lower()

            # Stop if it does not match the current character of the
            # desired word.
            if o != c:
                result = None
                break
            self.next()

        if fallback and not result:
            self.cursor = cursor

        return result

